#---------------------------------------------------------------------------------------
#   Supplementary File S3 for the manuscirpt 
#
#   "On the use of GBLUP and its extension for GWAS with additive and epistatic effects"
#---------------------------------------------------------------------------------------
#   R Code for generating the simulated phenotypic data 
#---------------------------------------------------------------------------------------

rm(list=ls())

#Function for phenotypic simulation.
#pve: a vector of PVE.
#ratio: a vector of Va/Vaa ratio.
#G: additive kinship matrix.
#H: AxA kinship matrix.
PheSim <- function(pve, ratio, G, H){
  sigma_a <- pve * (ratio/(1+ratio))
  sigma_aa <- pve * (1/(1+ratio))
  sd_a <- sqrt(sigma_a)
  sd_aa <- sqrt(sigma_aa)
  
  #Additive effect.
  eig.K <- eigen(G)
  u <- eig.K$vectors
  d <- sqrt(diag(eig.K$values))
  V <- u %*% d
  y_a <- rnorm(ngen, 0, sd_a)
  y_a <- V %*% y_a
  va <- var(y_a)
  
  #AxA epistatic effects.
  eig.K <- eigen(H)
  u <- eig.K$vectors
  d <- sqrt(diag(eig.K$values))
  V <- u %*% d
  y_aa <- rnorm(ngen, 0, sd_aa)
  y_aa <- V %*% y_aa
  vaa <- var(y_aa)
  
  #Residual effects.
  vg <- var(y_a + y_aa)
  sigma_e <- (1/pve - 1) * vg
  y_e <- rnorm(ngen, 0, sqrt(sigma_e))
  
  #Phenotypic value.
  y <- y_a + y_aa + y_e
  vp <- var(y)
  
  return(list(Phe=y, Vars=c(sigma_a, sigma_aa, sigma_e, va, vaa, vg, vp)))
}


#Read into genotypes.
GNdata <- read.table("File_S5.txt", header=TRUE)
GNdata <- as.matrix(GNdata)
alfreq <- apply(GNdata,2,mean)/2
ngen <- nrow(GNdata)
nmar <- ncol(GNdata)
alf <- alfreq

#Calculate the kinship matrices 
P <- matrix(rep(2*alf, ngen), nrow=ngen, byrow=TRUE)
W <- GNdata-P
c <- 2 * sum(alf*(1-alf))
Z <- W / sqrt(c)
G <- Z %*% t(Z)
H <- 0.5 * (G*G - (Z*Z) %*% (t(Z)*t(Z)))

#Making the kinship matirces positive-definite.
eigG <- eigen(G)
# eigG$values
d <- eigG$values
U <- eigG$vectors
zeropos1 <- which(d < 0.00001)
d0 <- d
if (length(zeropos1) > 0) {
  d0[zeropos1] <- 10^(floor(log10(min(d[-zeropos1]))))
}
G0 <- U%*%diag(d0)%*%t(U)

eigH <- eigen(H)
c <- eigH$values
V <- eigH$vectors
zeropos2 <- which(c < 0.00001)
c0 <- c
if (length(zeropos2) > 0) { 
  c0[zeropos1] <- 10^(floor(log10(min(c[-zeropos2]))))
}
H0 <- V%*%diag(c0)%*%t(V)

#Phenotye simulation.
#PVE:
pves <- seq(0.1, 0.9, 0.1)
#Addtive/Epistatic variance ratio.
ratios <- c(0.25, 0.5, 1, 2, 4, 8, 16)
var.res <- NULL
rep.times <- c(20)
out.dir <- c("simulated_phes")
system(paste0("mkdir ", out.dir))
sim.out <- NULL
for(i.x in 1:length(pves)){
  pve <- pves[i.x]
  for(i.y in 1:length(ratios)){
    ratio <- ratios[i.y]
    for(i in 1:rep.times){
      #Simulate phenotype.
      y.list <- PheSim(pve, ratio, G0, H0)
      
      #Output the phenotype.
      y.frame <- data.frame(StrainID=row.names(GNdata), Phe=y.list[[1]])
      y.name <- paste0("rice.phe_p", pve, "_r", ratio, "_", i, ".txt")
      write.table(y.frame, paste0(out.dir, "/", y.name), append=F, quote=F, sep="\t", col.names=T, row.names=F)
      
      #Output the parameters of the simulated phenotype.
      y.par <- c(pve, ratio, i, y.list[[2]])
      sim.out <- rbind(sim.out, y.par)
    }
  }
}
colnames(sim.out) <- c("pve", "ratio", "NO", "sigma_a", "sigma_aa", "sigma_e", "VarA", "VarAA", "VarG", "VarP")
write.table(sim.out, c("rice.simulated.phenotypes.inf.txt"), append=F, quote=F, sep="\t", col.names=T, row.names=F)

